require File.expand_path(File.dirname(__FILE__) + '/../spec_helper')

describe Pokan::Entity do
  context 'when comparing key-values' do
    before :all do
      Pokan::Connection.redis.flushall
      @entity = Pokan::Entity.new
      @entity.id = '1'
      @entity.store(:k1, "v1")
      @entity.store(:k2, 5.0)
    end

    it 'should return true if the internal key have the same value' do
      @entity.match?(k1: 'v1').should be_true
    end

    it 'should return false if the internal key doesnt have the same value' do
      @entity.match?(k1: 'v2').should_not be_true
    end

    it 'should return true if the internal key is one of the array' do
      @entity.match?(k1: ['v1', 'v2']).should be_true
    end

    it 'should return false if the internal key isnt one of the array' do
      @entity.match?(k1: ['v3', 'v2']).should_not be_true
    end

    it 'should return true if the value for a given key is in the range' do
      @entity.match?(k2: { min: 3.0, max: 10.0 }).should be_true
    end

    it 'should return true if the value for a given key is in the range' do
      @entity.match?(k2: { min:6.0, max: 10.0 }).should_not be_true
    end
  end

  context 'when storing a value' do
    before :all do
      Pokan::Connection.redis.flushall
      @entity = Pokan::Entity.new
      @entity.id = '1'
      @entity.store(:test, 'OK')
      @entity.save
      @entity.reload
    end

    it 'should store a new value for a key' do
      @entity.value(:test).should == 'OK'
    end

    it 'should store a timestamp for a key' do
      @entity.timestamp(:test).should < Time.now
    end

    it "should update the timestamp of an updated key" do
      new = @entity.timestamp(:status)
      @entity.store(:local, 'somewhere')
      @entity.save
      @entity.reload
      newer = @entity.timestamp(:local)
      newer.to_f.should > new.to_f
    end
  end

  context 'when retrieving a value' do
    before :all do
      Pokan::Connection.redis.flushall
      @entity = Pokan::Peer.new
      @entity.id = '1'
      @entity.store(:test, 'OK')
      @entity.save
      @entity.reload
    end

    it 'should get the value for a given key' do
      @entity.value(:test).should == 'OK'
    end

    it 'should return nil if a value doesn\'t exist' do
      @entity.value(:other_test).should be_nil
    end

    it 'should return zero to timestamp if a value doesn\'t exist' do
      @entity.timestamp(:other_test).should == Time.at(0)
    end
  end

  context 'when merging new data to a entity' do
    before :each do
      Pokan::Connection.redis.flushall
      @entity = Pokan::Entity.new
      @entity.id = '1'
      @entity.store(:local, 'GREAT')
      @entity.save
      @entity.reload
    end

    it 'should merge with a new set preserving newer values' do
      @entity.merge({
        local: { value: 'GREATER', timestamp: Time.now.to_f + 2000 }
      })
      @entity.value(:local).should == 'GREATER'
    end

    it 'should merge with a new set discarding older values' do
      @entity.merge({
        local: {value: 'NOT GREAT', timestamp: 50}
      })
      @entity.value(:local).should == 'GREAT'
    end
  end

  context 'when getting data for gossip' do
    before :all do
      Pokan::Connection.redis.flushall
      @entity = Pokan::Entity.new
      @entity.id = '1'
      @entity.store(:foo, 'bar')
      @entity.store(:local, 'OK')
      @entity.save
      @entity.reload
    end

    it 'should return all key/timestamps of a peer' do
      @entity.digest.keys.should =~ [:local, :foo]
    end

    it 'should return all older keys of a peer' do
      digest = {local: 0, foo: (Time.now.to_f + 1000).to_s}
      @entity.older(digest).should == [:foo]
    end

    it 'should return all newer key/value/timestamps of an peer' do
      digest = {local: 0, foo: (Time.now.to_f + 1000).to_s}
      @entity.newer(digest)[:local][:value].should_not == nil
    end

    it 'should return all the peer\'s keys' do
      @entity.keys.should =~ [:foo, :local]
    end

    it 'should return the keys, values and timestamps of all required keys' do
      @entity.values([:local])[:local][:value].should == 'OK'
    end
  end
end
